@@ -8,7 +8,7 @@
 
 #include "procore.h"
 #include "prostate.h"
-#include "pstack.h"
+#include "tagstack.h"
 #include "pbuffer.h"
 #include "proscope.h"
 #include "expr.h"
@@ -21,7 +21,8 @@
 #define HTML_TEMPLATE_TAG_IF      4
 #define HTML_TEMPLATE_TAG_ELSE    5
 #define HTML_TEMPLATE_TAG_UNLESS  6
-#define HTML_TEMPLATE_LAST_TAG_USED  6
+#define HTML_TEMPLATE_TAG_ELSIF   7
+#define HTML_TEMPLATE_LAST_TAG_USED  7
 
 #if defined(__GNUC__) && !(defined(PEDANTIC))
 #define INLINE inline
@@ -32,13 +33,13 @@
 static 
 const char* const tagname[]={
     "Bad or unsupported tag", /* 0 */
-    "var", "include", "loop", "if", "else", "unless"
+    "var", "include", "loop", "if", "else", "unless", "elsif"
 };
 
 static 
 const char* const TAGNAME[]={
     "Bad or unsupported tag", /* 0 */
-    "VAR", "INCLUDE", "LOOP", "IF", "ELSE", "UNLESS"
+    "VAR", "INCLUDE", "LOOP", "IF", "ELSE", "UNLESS", "ELSIF"
 };
 
 /* max offset to ensure we are not out of file when try <!--/  */
@@ -102,7 +103,7 @@ get_loop_context_vars_value (struct tmplpro_state *state, PSTRING name) {
   static char buffer[20]; /* for snprintf %d */
   int loop;
   PSTRING retval={NULL,NULL};
-  if (curScopeLevel(&state->var_scope_stack)>0 
+  if (curScopeLevel(&state->param->var_scope_stack)>0 
       && name.endnext-name.begin>4
       && '_'==*(name.begin)
       && '_'==*(name.begin+1)
@@ -110,7 +111,7 @@ get_loop_context_vars_value (struct tmplpro_state *state, PSTRING name) {
     /* we can meet loop variables here -- try it first */
     /* length of its name >4 */
     /* __first__ __last__ __inner__ __odd__ __counter__ */
-    struct ProLoopState* currentScope = getCurrentScope(&state->var_scope_stack);
+    struct ProLoopState* currentScope = getCurrentScope(&state->param->var_scope_stack);
     PSTRING shiftedname; /* (PSTRING) {name.begin+2,name.endnext} */
     shiftedname.begin=name.begin+2;
     shiftedname.endnext=name.endnext;
@@ -277,12 +278,12 @@ _tmpl_log_state (struct tmplpro_state *state, int level)
 static
 ABSTRACT_VALUE* walk_through_nested_loops (struct tmplpro_state *state, PSTRING name) {
   int PrevHash;
-  struct ProLoopState* currentScope = getCurrentScope(&state->var_scope_stack);
+  struct ProLoopState* currentScope = getCurrentScope(&state->param->var_scope_stack);
   ABSTRACT_VALUE* valptr= state->param->getAbstractValFuncPtr(currentScope->param_HV, name);
   if ((0==state->param->global_vars) || (valptr)) return valptr;
-  PrevHash=curScopeLevel(&state->var_scope_stack)-1;
+  PrevHash=curScopeLevel(&state->param->var_scope_stack)-1;
   while (PrevHash>=0) {
-    valptr=state->param->getAbstractValFuncPtr(getScope(&state->var_scope_stack, PrevHash)->param_HV,name);
+    valptr=state->param->getAbstractValFuncPtr(getScope(&state->param->var_scope_stack, PrevHash)->param_HV,name);
     if (valptr!=NULL) return valptr;
     PrevHash--;
   }
@@ -313,7 +314,7 @@ tag_handler_var (struct tmplpro_state *state, PSTRING name, PSTRING defvalue, in
     varvalue=get_variable_value(state, name);
   if (debuglevel>=TMPL_LOG_DEBUG) {
       if (varvalue.begin!=NULL) {
-      tmpl_log(state,TMPL_LOG_DEBUG,"variable value = %.*s\n",varvalue.endnext-varvalue.begin,varvalue.begin);
+	tmpl_log(state,TMPL_LOG_DEBUG,"variable value = %.*s\n",(int)(varvalue.endnext-varvalue.begin),varvalue.begin);
     } else {
       tmpl_log(state,TMPL_LOG_DEBUG,"variable value = UNDEF\n");
     }
@@ -395,7 +396,7 @@ static
 void 
 tag_handler_if (struct tmplpro_state *state, PSTRING name)
 {
-  struct pstack_entry iftag;
+  struct tagstack_entry iftag;
   iftag.tag=HTML_TEMPLATE_TAG_IF;
   iftag.vcontext=state->is_visible;
   iftag.position=state->cur_pos; /* unused */
@@ -406,7 +407,7 @@ tag_handler_if (struct tmplpro_state *state, PSTRING name)
     iftag.value=0;
     state->is_visible=0;
   }
-  pstack_push(&(state->tag_stack), iftag);
+  tagstack_push(&(state->tag_stack), iftag);
   if (debuglevel>3) tmpl_log(state,TMPL_LOG_DEBUG2,"tag_handler_if:visible context =%d value=%d ",iftag.vcontext,iftag.value);
 }
 
@@ -414,17 +415,17 @@ static
 void 
 tag_handler_unless (struct tmplpro_state *state, PSTRING name)
 {
-  struct pstack_entry iftag;
+  struct tagstack_entry iftag;
   iftag.tag=HTML_TEMPLATE_TAG_UNLESS;
   iftag.vcontext=state->is_visible;
   iftag.position=state->cur_pos; /* unused */
   if (is_var_true(state,name)) {
     iftag.value=0;
-    pstack_push(&(state->tag_stack), iftag);
+    tagstack_push(&(state->tag_stack), iftag);
     state->is_visible=0;
   } else {
     iftag.value=1;
-    pstack_push(&(state->tag_stack), iftag);
+    tagstack_push(&(state->tag_stack), iftag);
     /* state->is_visible is not touched */
   }
   if (debuglevel>3) tmpl_log(state,TMPL_LOG_DEBUG2,"unless:visible context =%d value=%d ",iftag.vcontext,iftag.value);
@@ -435,7 +436,7 @@ INLINE
 int
 test_stack (int tag)
 {
-  //  return (pstack_notempty(&(state->tag_stack)) && (pstack_head(&(state->tag_stack))->tag==tag));
+  //  return (tagstack_notempty(&(state->tag_stack)) && (tagstack_top(&(state->tag_stack))->tag==tag));
   return 1;
 }
 
@@ -444,10 +445,10 @@ void
 tag_stack_debug  (struct tmplpro_state *state, int stack_tag_type)
 {
   if (stack_tag_type) {
-    if (pstack_notempty(&(state->tag_stack))) {
-      struct pstack_entry iftag=pstack_top(&(state->tag_stack));
-      if (iftag.tag!=stack_tag_type) {
-	tmpl_log(state,TMPL_LOG_ERROR, "ERROR: tag mismatch with %s\n",TAGNAME[iftag.tag]);
+    if (tagstack_notempty(&(state->tag_stack))) {
+      struct tagstack_entry* iftag=tagstack_top(&(state->tag_stack));
+      if (iftag->tag!=stack_tag_type) {
+	tmpl_log(state,TMPL_LOG_ERROR, "ERROR: tag mismatch with %s\n",TAGNAME[iftag->tag]);
       }
     } else {
       tmpl_log(state,TMPL_LOG_ERROR, "ERROR: opening tag %s not found\n",TAGNAME[stack_tag_type]);
@@ -459,12 +460,12 @@ static
 void 
 tag_handler_closeif (struct tmplpro_state *state)
 {
-  struct pstack_entry iftag;
+  struct tagstack_entry iftag;
   if (! test_stack(HTML_TEMPLATE_TAG_IF)) {
     tag_stack_debug(state,HTML_TEMPLATE_TAG_IF);
     return;
   }
-  iftag=pstack_pop(&(state->tag_stack));
+  iftag=tagstack_pop(&(state->tag_stack));
   if (0==state->is_visible) state->last_processed_pos=state->cur_pos;
   state->is_visible=iftag.vcontext;
 }
@@ -473,12 +474,12 @@ static
 void 
 tag_handler_closeunless (struct tmplpro_state *state)
 {
-  struct pstack_entry iftag;
+  struct tagstack_entry iftag;
   if (! test_stack(HTML_TEMPLATE_TAG_UNLESS)) {
     tag_stack_debug(state,HTML_TEMPLATE_TAG_UNLESS);
     return;
   }
-  iftag=pstack_pop(&(state->tag_stack));
+  iftag=tagstack_pop(&(state->tag_stack));
   if (0==state->is_visible) state->last_processed_pos=state->cur_pos;
   state->is_visible=iftag.vcontext;
 }
@@ -487,20 +488,48 @@ static
 void 
 tag_handler_else (struct tmplpro_state *state)
 {
-  struct pstack_entry iftag;
+  struct tagstack_entry* iftag;
   if (! test_stack(HTML_TEMPLATE_TAG_IF) && 
       ! test_stack(HTML_TEMPLATE_TAG_UNLESS)) {
     tag_stack_debug(state,HTML_TEMPLATE_TAG_ELSE);
     return;
   }
-  iftag=pstack_top(&(state->tag_stack));
+  iftag=tagstack_top(&(state->tag_stack));
   if (0==state->is_visible) state->last_processed_pos=state->cur_pos;
-  if (iftag.value) {
+  if (iftag->value) {
     state->is_visible=0;
-  } else if (1==iftag.vcontext) {
+  } else if (1==iftag->vcontext) {
     state->is_visible=1;
   }
-  if (debuglevel>3) tmpl_log(state,TMPL_LOG_DEBUG2,"else:(pos %td) visible:context =%d, set to %d ",iftag.position - state->top,iftag.vcontext,state->is_visible);
+  if (debuglevel>3) tmpl_log(state,TMPL_LOG_DEBUG2,"else:(pos %td) visible:context =%d, set to %d ",iftag->position - state->top,iftag->vcontext,state->is_visible);
+}
+
+static 
+void 
+tag_handler_elsif (struct tmplpro_state *state, PSTRING name)
+{
+  struct tagstack_entry *iftag;
+  if (! test_stack(HTML_TEMPLATE_TAG_IF) && 
+      ! test_stack(HTML_TEMPLATE_TAG_UNLESS)) {
+    tag_stack_debug(state,HTML_TEMPLATE_TAG_ELSIF);
+    return;
+  }
+  iftag=tagstack_top(&(state->tag_stack));
+  if (0==state->is_visible) state->last_processed_pos=state->cur_pos;
+  if (iftag->value) {
+    state->is_visible=0;
+  } else if (1==iftag->vcontext) {
+    /* test only if vcontext==true; if the whole tag if..endif itself is invisible, skip the is_var_true test */
+    /*TODO: it is reasonable to skip is_var_true test in if/unless too */
+    if (is_var_true(state,name)) {
+      iftag->value=1;
+      state->is_visible=1;
+    } else {
+      iftag->value=0;
+      state->is_visible=0;
+    }
+  }
+  if (debuglevel>3) tmpl_log(state,TMPL_LOG_DEBUG2,"elsif:(pos %td) visible:context =%d, set to %d ",iftag->position - state->top,iftag->vcontext,state->is_visible);
 }
 
 static 
@@ -509,9 +538,9 @@ next_loop (struct tmplpro_state* state) {
 #ifdef DEBUG
   tmpl_log(state,TMPL_LOG_DEBUG2,"next_loop:before NextLoopFuncPtr\n");
 #endif
-  struct ProLoopState* currentScope = getCurrentScope(&state->var_scope_stack);
+  struct ProLoopState* currentScope = getCurrentScope(&state->param->var_scope_stack);
   if (++currentScope->loop >currentScope->maxloop || !(state->param->NextLoopFuncPtr)(currentScope)) {
-    popScope(&state->var_scope_stack);
+    popScope(&state->param->var_scope_stack);
     return 0;
   } else return 1;
 }
@@ -520,7 +549,7 @@ static
 void 
 tag_handler_loop (struct tmplpro_state *state, PSTRING name)
 {
-  struct pstack_entry iftag;
+  struct tagstack_entry iftag;
   iftag.tag=HTML_TEMPLATE_TAG_LOOP;
   iftag.vcontext=state->is_visible;
   iftag.value=0;
@@ -528,7 +557,7 @@ tag_handler_loop (struct tmplpro_state *state, PSTRING name)
 #ifdef DEBUG
   tmpl_log(state,TMPL_LOG_DEBUG2,"tag_handler_loop:before InitLoopFuncPtr\n");
 #endif
-  if (state->is_visible && (*state->param->InitLoopFuncPtr)(&state->var_scope_stack,name) && next_loop(state)) {
+  if (state->is_visible && (*state->param->InitLoopFuncPtr)(&state->param->var_scope_stack,name) && next_loop(state)) {
     iftag.value=1; /* the loop is non - empty */
   } else {
     /* empty loop is equal to <if false> ... </if> */
@@ -537,27 +566,28 @@ tag_handler_loop (struct tmplpro_state *state, PSTRING name)
 #ifdef DEBUG
   tmpl_log(state,TMPL_LOG_DEBUG2,"tag_handler_loop:after InitLoopFuncPtr\n");
 #endif
-  pstack_push(&(state->tag_stack), iftag);
+  tagstack_push(&(state->tag_stack), iftag);
 }
 
 static 
 void 
 tag_handler_closeloop (struct tmplpro_state *state)
 {
-  struct pstack_entry iftag;
+  struct tagstack_entry* iftag_ptr;
   if (! test_stack(HTML_TEMPLATE_TAG_LOOP)) {
     tag_stack_debug(state,HTML_TEMPLATE_TAG_LOOP);
     return;
   }
-  iftag=pstack_top(&(state->tag_stack));
-  if (iftag.value==1 && next_loop(state)) {
+  iftag_ptr=tagstack_top(&(state->tag_stack));
+  if (iftag_ptr->value==1 && next_loop(state)) {
     /* continue loop */
-    state->cur_pos=iftag.position;
-    state->last_processed_pos=iftag.position;
+    state->cur_pos=iftag_ptr->position;
+    state->last_processed_pos=iftag_ptr->position;
     return;
   } else {
     /* finish loop */
-    iftag=pstack_pop(&(state->tag_stack));
+    struct tagstack_entry iftag;
+    iftag=tagstack_pop(&(state->tag_stack));
     state->is_visible=iftag.vcontext;
     state->last_processed_pos=state->cur_pos;
   }
@@ -792,7 +822,7 @@ process_tmpl_tag(struct tmplpro_state *state)
 	  escape = HTML_TEMPLATE_OPT_ESCAPE_NO;
 	  break;
 	default:
-	  tmpl_log(state,TMPL_LOG_ERROR, " unsupported value of ESCAPE=%.*s\n",OptEscape.endnext-OptEscape.begin,OptEscape.begin);
+	  tmpl_log(state,TMPL_LOG_ERROR, " unsupported value of ESCAPE=%.*s\n",(int)(OptEscape.endnext-OptEscape.begin),OptEscape.begin);
 	}
       }
       tag_handler_var(state,OptName,OptDefault,escape);
@@ -800,6 +830,7 @@ process_tmpl_tag(struct tmplpro_state *state)
     case HTML_TEMPLATE_TAG_IF:	tag_handler_if(state,OptName);	break;
     case HTML_TEMPLATE_TAG_UNLESS:	tag_handler_unless(state,OptName);break;
     case HTML_TEMPLATE_TAG_ELSE:	tag_handler_else(state);	break;
+    case HTML_TEMPLATE_TAG_ELSIF:	tag_handler_elsif(state,OptName);break;
     case HTML_TEMPLATE_TAG_LOOP:	tag_handler_loop(state,OptName);break;
     case HTML_TEMPLATE_TAG_INCLUDE:	tag_handler_include(state,OptName,OptDefault);break;
     default:	tag_handler_unknown(state);break;
@@ -817,11 +848,11 @@ process_state (struct tmplpro_state * state)
   flag is_tag_closed;
   flag is_tag_commented;
   char* last_safe_pos=state->next_to_end-TAG_WIDTH_OFFSET;
-  pstack_init(&(state->tag_stack));
+  if (debuglevel) tmpl_log(NULL,TMPL_LOG_DEBUG,"process_state:initiated\n");
+  tagstack_init(&(state->tag_stack));
   pbuffer_init(&(state->str_buffer));
-  Scope_init(&state->var_scope_stack);
-  // hack (until param and state will merge ???)
-  setRootScope(&state->var_scope_stack,state->param->rootHV);
+  pbuffer_init(&(state->expr_pusharg_buffer));
+  Scope_init_root(&state->param->var_scope_stack,state->param->rootHV);
 
 
   while (state->cur_pos < last_safe_pos) {
@@ -849,9 +880,10 @@ process_state (struct tmplpro_state * state)
   }
   (state->param->WriterFuncPtr)(state->last_processed_pos,state->next_to_end);
 
-  Scope_free(&state->var_scope_stack);
+  Scope_free(&state->param->var_scope_stack);
+  pbuffer_free(&(state->expr_pusharg_buffer));
   pbuffer_free(&(state->str_buffer));
-  pstack_free(&(state->tag_stack));
+  tagstack_free(&(state->tag_stack));
   if (debuglevel) tmpl_log(NULL,TMPL_LOG_DEBUG,"process_state:finished\n");
 }
 
@@ -922,7 +954,7 @@ void
 tmplpro_procore_init()
 {
   /* to save time on malloc/free at each included template */
-  /* pstack_selftest(); */
+  /* tagstack_selftest(); */
 }
 
 void 
@@ -957,7 +989,7 @@ tmplpro_param_free(struct tmplpro_param* param)
   free(param);
 }
 
-#include "pstack.inc"
+#include "tagstack.inc"
 
 /*
  * Local Variables:
